import logging

from requests.api import post
from requests.exceptions import RequestException


class TemporaryError(Exception):
    """
    Exception class to indicate there was recoverable, temporary error
    while communicating with the AFM.
    """
    def __init__(self, what):
        self.what = what

    def __str__(self):
        return str(self.what)


class FatalError(Exception):
    """
    Exception class to indicate there was unrecoverable error, most likely
    due to discontinued legacy protocol support.
    """
    def __init__(self, what):
        self.what = what

    def __str__(self):
        return str(self.what)


class Communicator(object):
    """
    Class to handle afm communication and to provide
    simple error handling interface for the agent application.
    """

    def __init__(self, afm_url):
        self.logger = logging.getLogger("Communicator")
        self.afm_url = afm_url

    def _post(self, path, json):
        """
        Utility to simplify requests to afm and handle (translate) exceptions
        :param path:
        :param json:
        :return:
        """

        try:
            r = post(
                self.afm_url + path,
                json=json
            )
            if 400 <= r.status_code < 500:
                raise FatalError('AFM responded with client error status: {0}'.format(r.status_code))
            elif 500 <= r.status_code < 600:
                raise TemporaryError('AFM responded with server error status: {0}'.format(r.status_code))
            return r
        except RequestException as e:
            raise TemporaryError(e)

    def request_tasks(self, **kwargs):
        """
        Request tasks from the afm.

        :param kwargs: Named parameters to pass in task request JSON.
        :return: parsed response JSON
        """

        request_data = {
            "protocol": 1
        }
        request_data.update(kwargs)

        logging.getLogger('HTTP').debug("Requesting for tasks: %s", request_data)
        r = self._post('/tasks', json=request_data)

        try:
            response_data = r.json()
            return response_data
        except Exception as e:
            raise FatalError('Failed to parse response JSON: {0}'.format(e))

    def post_result(self, task_id, task_data=None, task_error=None):
        """
        Post task result to the afm.
        Will handle temporary errors and retry post attempts until succeeds.

        :param task_id: id of the related task
        :param task_data: result data of the task if execution succeeded
        :param task_error: error message when task execution failed
        """

        request_data = {
            "protocol": 1,
            "task_id": task_id
        }
        if task_data:
            request_data["task_data"] = task_data
        elif task_error:
            request_data["task_error"] = task_error

        logging.getLogger('HTTP').debug("Posting results for task '%s': %s", task_id, request_data)

        uploaded = False
        while not uploaded:
            try:
                uploaded = self._post('/tasks/response', json=request_data).status_code == 200
            except TemporaryError as e:
                self.logger.warning("Temporary error while posting results: %s", e)



import logging
import uuid
import time
import argparse
import importlib

from slamon_agent import timeutil
from slamon_agent.executor import Executor
from slamon_agent.communication import TemporaryError
from slamon_agent.communication import Communicator
from slamon_agent.handlers import TaskHandler


class Agent(object):
    """
    Agent class presents an instance of agent application.
    """

    def __init__(self, afm_url, default_wait=60, max_tasks=2, name='Python Agent 1.0', agent_uuid=None):

        self.afm = Communicator(afm_url)
        self.max_tasks = max_tasks
        self.default_wait = default_wait
        self.name = name
        self.uuid = agent_uuid if agent_uuid else str(uuid.uuid1())
        self._run = True

    def exit(self):
        """
        Signal agent to exit. After issuing exit, agent will not make further task requests,
        but will wait until all currently processed tasks finish.
        """
        self._run = False

    def run(self):
        """
        The "main function" of the agent, looping the claim & execute tasks flow.
        """

        with Executor(self.max_tasks) as executor:
            while self._run:

                wait_time = self.default_wait
                min_wait_time = 1

                # request for tasks
                try:
                    task_response = self.afm.request_tasks(
                        agent_id=self.uuid,
                        agent_name=self.name,
                        agent_time=timeutil.format(timeutil.now()),
                        agent_capabilities=TaskHandler.list_all(),
                        max_tasks=executor.available_executors()
                    )
                    if 'tasks' in task_response:
                        for task_data in task_response['tasks']:
                            executor.submit_task(task_data,
                                                 lambda *args, **kwargs: self.afm.post_result(*args, **kwargs))
                    if 'return_time' in task_response:
                        return_time = timeutil.parse(task_response['return_time'])
                        wait_time = max(min_wait_time, (return_time - timeutil.now()).total_seconds())
                except TemporaryError as e:
                    logging.getLogger("Agent").error("An error occurred while claiming tasks: %s", e)

                time.sleep(wait_time)


def _import_module(module_name, package=None):
    """
    Recursively load modules to search for task handlers.
    """
    m = importlib.import_module(module_name, package=package)
    if hasattr(m, '__all__'):
        for sub_module_name in m.__all__:
            _import_module('.' + sub_module_name, module_name)


def main():
    """
    Entry point for the agent script.
    """
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--url', action='store', required=True,
                        help='Coordinator URL')
    parser.add_argument('-l', '--load', action='append',
                        help='Load handlers from specified module or package e.g. slamon_agent.handlers')
    parser.add_argument('-w', '--default-wait', type=int, default=60,
                        help='Seconds to wait before reconnection after connection failure.')
    parser.add_argument('-x', '--num-executors', type=int, default=2,
                        help='Number of concurrent task executors.')
    args = parser.parse_args()

    # import defined modules to load defined handlers
    if args.load:
        for module in args.load:
            _import_module(module)

    agent = Agent(
        args.url,
        default_wait=args.default_wait,
        max_tasks=args.num_executors
    )
    agent.run()
    return 0


if __name__ == "__main__":
    main()



from datetime import datetime

from dateutil import parser, tz


def now():
    """
    Utility to get current datetime with local timezone data.
    :return: datetime object with current time and timezone
    """

    return datetime.now(tz=tz.tzlocal())


def parse(datetime_str):
    """
    Utility to parse datetime and timezone from string formatted as specified in AFM REST API.
    :param datetime_str: datetime string with timezone data e.g. '2012-04-23T18:25:43.511Z'
    :return: datetime object with timezone data
    """

    datetime = parser.parse(datetime_str)
    if not datetime.tzinfo:
        raise Exception("No time zone info in date time input!")
    return datetime


def format(dt):
    """
    Utility to format date as specified in AFM REST API.
    :param dt: datetime object to format
    :return:
    """
    return dt.isoformat()



from slamon_agent.handlers import TaskHandler
import requests


@TaskHandler('url_http_status', 1)
def url_http_status_task_handler(input_params):
    """
    Task for checking web page availability and returning status code if page is available
    """
    url = input_params['url']

    response = requests.get(url)
    return {'status': response.status_code}



from slamon_agent.handlers import TaskHandler


@TaskHandler("wait", 1)
def wait_task_handler(input_params):
    """
    Agent for prototyping. Will wait for approximately the specified
    timeout, with slight random factor.
    """
    import time
    import random

    timeout = float(input_params['time']) - 0.5 + random.random()
    time.sleep(timeout)
    return {'time': timeout}



from concurrent.futures import ThreadPoolExecutor
import logging
import threading

from slamon_agent.handlers import TaskHandler


class Executor(object):
    """
    Task executor pool to run task handlers asynchronously.
    To provide clean shutdown, Executor is a python context manager and
    performs shutdown routines on context exit.

    Example:
        Executing tasks in existing list ``task_list`` with two concurrent executors:

            from slamon.agent.executor import Executor

            def result_callback(task_id, task_data=None, task_error=none):
                if task_data:
                    print('Task {0} finished with data: {1}'.format(task_id, task_data))
                else:
                    print('Task {0} finished with error: {1}'.format(task_id, task_error))

            with Executor(max_executors=2) as executor:
                for task in task_list:
                    executor.submit_task(task, result_callback)

    """

    def __init__(self, max_executors=2):
        self._max_executors = max_executors
        self._active_executors = 0
        self._lock = threading.Lock()
        self._logger = logging.getLogger('Executor')

    def __enter__(self):
        """
        Context manager entry function to start thread pool.
        """
        self._logger.debug('Starting executor with %d threads', self._max_executors)
        self._thread_pool = ThreadPoolExecutor(self._max_executors)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager exit function to shutdown thread pool.
        """
        self._logger.debug('Stopping executor')
        self._thread_pool.shutdown()
        self._thread_pool = None

    def _run_task(self, task, callback):
        try:
            self._logger.debug('Creating task handler for task %s', task['task_id'])
            handler = TaskHandler.create(task['task_type'], task['task_version'])

            self._logger.debug('Executing task handler for task %s', task['task_id'])
            result = handler(task['task_data'])

            self._logger.debug('Task executed successfully: %s', task['task_id'])
            callback(task['task_id'], task_data=result)

        except Exception as e:
            self._logger.warning('Execution of task %s raised an exception: %s', task['task_id'], str(e))
            callback(task['task_id'], task_error=str(e))

        finally:
            with self._lock:
                self._active_executors -= 1

    def submit_task(self, task, callback):
        """
        Queue task for asynchronous execution. The callback will receive
        task id as first positional parameter and either task_data or task_error
        named parameter describing the output.

        :param task: dictionary describing the task
        :param callback: callable that will receive results
        """
        with self._lock:
            self._thread_pool.submit(lambda: self._run_task(task, callback))
            self._active_executors += 1

    def available_executors(self):
        """
        Get number of available executors. (max executors - active executors)

        :return: number of available executors
        """
        with self._lock:
            return max(self._max_executors - self._active_executors, 0)



class TaskHandler(object):
    """
    TaskHandler is a decorator to register a callable as a task handler for the agent.
    """

    _handlers = {}

    def __init__(self, name, version):
        self.name = name
        self.version = version

    def __call__(self, handler):
        TaskHandler.register(handler, self.name, self.version)
        return handler

    @classmethod
    def register(cls, callable, name, version):
        """
        Utility to register any callable as task handler.
        :param callable: Callable taking one positional argument (input data) and returning dictionary of output params.
        :param name: Name of the task type
        :param version: Version of the supported task type
        :return:
        """
        import logging

        logging.getLogger("TaskHandler").debug("Registering handler: %s", name)
        TaskHandler._handlers[name] = {
            'version': version,
            'function': callable
        }

    @classmethod
    def create(cls, name, version):
        """
        Get corresponding handler for task.

        :param name: the name of the task
        :param version: the version of the task
        :return: handler is suitable is found, None otherwise
        """
        handler = TaskHandler._handlers[name]
        if handler['version'] == version:
            return handler['function']

    @classmethod
    def list_all(cls):
        """
        List all registered task handlers as dictionary consisting of task names as keys and dictionaries as values.

        :return: dictionary of registered task handlers
        """
        return {name: {'version': data['version']} for name, data in TaskHandler._handlers.items()}



#!/usr/bin/env python

from setuptools import setup

setup(
    name='slamon-agent',
    version='1.0.0',
    description='Python implementation of a SLAMon Agent',
    url='https://github.com/SLAMon/slamon-python-agent',
    author='SLAMon',
    author_email='slamon.organization@gmail.com',
    license='Apache License v2.0',
    platforms=['Python 3.3+'],
    long_description='The implementation of SLAMon agent on Python 3.4. The agent connects to an Agent Fleet Manager, \
and executes tasks retrieved.\nAimed to be used on Unix platforms. The agent can be used either from a python script \
or via command line script.\nRead more from the repository for examples and instructions.',
    packages=[
        'slamon_agent',
        'slamon_agent.handlers',
    ],
    install_requires=[
        'python_dateutil>= 2.4.2, <3.0',
        'requests>=2.5.4.1, <3.0'
    ],
    entry_points={
        'console_scripts': [
            'slamon-agent = slamon_agent.agent:main'
        ]
    },
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: System :: Monitoring'
    ]
)
